# TP 15 correction : Matrices
# BCPST1B 2023-2024
# Lycée Hoche, Versailles
# L.-C. LEFÈVRE

#%% fonctions données
# exécuter une fois pour toute la cellule

from random import randint
import numpy as np # pour tricher

# matrice nulle de taille (n, p)
def matrice_nulle(n, p):
    return [[0 for _ in range(p)] for _ in range(n)]

# matrice de coefficients au hasard entre -9 et 9, de taille (n, p)
def matrice_hasard(n, p):
    return [[randint(-9, 9) for _ in range(p)] for _ in range(n)]

# affiche la matrice, bien mise en forme, en trichant
def matrice_print(A):
    print(np.array(A))

#%% exercice 1

def taille(A):
    n = len(A)
    p = len(A[0])
    return (n, p)

#%% exercice 2

def identité(n):
    A = matrice_nulle(n, n)
    for i in range(n):
        A[i][i] = 1
    return A

# test
A = identité(5)
matrice_print(A)

#%% exercice 3

def diagonale(L):
    n = len(L)
    A = matrice_nulle(n, n)
    for i in range(n):
       A[i][i] = L[i]
    return A

# test
L = [1, 3, 5, 7]
A = diagonale(L)
matrice_print(A)

#%% exercice 4

def somme(A, B):
    (n, p) = taille(A)
    (n2, p2) = taille(B)
    assert n == n2 and p == p2, "ça ne marche pas, revois ton cours"
    S = matrice_nulle(n, p)
    for i in range(n):
        for j in range(p):
            S[i][j] = A[i][j] + B[i][j]
    return S

# test
A = matrice_hasard(4, 5)
B = matrice_hasard(4, 5)
matrice_print(A)
matrice_print(B)
S = somme(A, B)
matrice_print(S)

#%% exercice 5

def produit_constante(A, a):
    (n, p) = taille(A)
    B = matrice_nulle(n, p)
    for i in range(n):
        for j in range(p):
            B[i][j] = a * A[i][j]
    return B

# test
A = matrice_hasard(4, 5)
a = randint(-9, 9)
matrice_print(A)
print(a)
B = produit_constante(A, a)
matrice_print(B)

#%% exercice 6

def coeff_produit(A, B, i, j):
    (n, q) = taille(A)
    (q2, p) = taille(B)
    assert q == q2, "aïe, tu devrais revoir ton cours"
    c = 0
    for k in range(q):
        c = c + A[i][k] * B[k][j]
    return c

def produit(A, B):
    (n, q) = taille(A)
    (q2, p) = taille(B)
    assert q == q2, "il faut VRAIMENT que tu revoies ton cours"
    P = matrice_nulle(n, p)
    for i in range(n):
        for j in range(p):
            P[i][j] = coeff_produit(A, B, i, j)
    return P

# test
A = matrice_hasard(2, 3)
B = matrice_hasard(3, 4)
matrice_print(A)
matrice_print(B)
P = produit(A, B)
matrice_print(P)

# bonus : triple boucle
def produit2(A, B):
    (n, q) = taille(A)
    (q2, p) = taille(B)
    assert q == q2
    P = matrice_nulle(n, p)
    for i in range(n):
        for j in range(p):
            for k in range(q):
                P[i][j] = P[i][j] + A[i][k] * B[k][j]
    return B

#%% exercice 8

# version récursive
def puissance(A, N):
    (n, p) = taille(A)
    assert n == p, "non mais sérieusement, encore ???"
    if N == 0:
        return identité(n)
    else:
        return produit(A, puissance(A, N-1))

# version itérative
def puissance_itératif(A, N):
    (n, p) = taille(A)
    assert n == p
    P = identité(n)
    for k in range(N):
        P = produit(A, P)
    return P

# test, voir cours et programme de colle
A = [[2, 1, 0], [0, 2, 1], [0, 0, 2]]
matrice_print(A)
for N in range(5):
    print(N)
    P = puissance(A, N)
    matrice_print(P)

#%% exercice 9

def est_diagonale(A):
    """ renvoie True si la matrice A est diagonale, False sinon """
    (n, p) = taille(A)
    if n != p:
        return False # les matrices diagonales sont carrées
    for i in range(n):
        for j in range(p):
            if i != j and A[i][j] != 0:
                return False
    return True

# test
A = [[1, 0, 0], [0, 3, 0], [0, 0, 5]]
matrice_print(A)
print(est_diagonale(A))
B = [[1, 2, 0], [0, 3, 0], [0, 0, 5]]
matrice_print(B)
print(est_diagonale(B))
C = [[0, 0, 0], [0, 3, 0], [0, 0, 5]]
matrice_print(C)
print(est_diagonale(C))
D = [[0, 0, 0], [-1, 3, 0], [0, 0, 5]]
matrice_print(D)
print(est_diagonale(D))
E = [[1, 0, 0], [0, 3, 0], [0, 1, 5]]
matrice_print(E)
print(est_diagonale(E))

#%% exercice 10

def est_triangulaire_supérieure(A):
    """ renvoie True si la matrice A est triangulaire supérieure, False sinon """
    (n, p) = taille(A)
    if n != p:
        return False # les matrices triangulaires sont carrées
    for i in range(n):
        for j in range(p):
            if i > j and A[i][j] != 0:
                return False
    return True

A = [[1, 2, 3], [0, 3, -2], [0, 0, 5]]
matrice_print(A)
print(est_triangulaire_supérieure(A))
B = [[0, 2, 1], [0, 3, -2], [0, 0, 5]]
matrice_print(B)
print(est_triangulaire_supérieure(B))
C = [[1, 2, 3], [-7, 3, 0], [0, 0, 5]]
matrice_print(C)
print(est_triangulaire_supérieure(C))
D = [[1, 0, 0], [0, 3, 4], [0, 2, 5]]
matrice_print(D)
print(est_triangulaire_supérieure(D))

#%% à vous !

# on pourra échelonner les matrices suivantes
A1 = [[1, 0, -1], [-2, -1, 4], [0, 1, -3]]
A2 = [[-1, 3, -1], [0, -2, 0], [2, -6, 2], [1, -4, 1]]
A3 = [[-1, -3, -4, -2, 4], [-1, 2, 4, 5, -1], [-1, -1, -1, 1, 2], [0, 0, 5, -5, 0], [-1, 0, 4, -1, 1], [1, 1, -1, 1, -2]]

#%% exercice 11

def echange(A, i, j):
    (n, p) = taille(A)
    for k in range(p):
        x = A[i][k]
        A[i][k] = A[j][k]
        A[j][k] = x

def combinaison(A, a, i, b, j):
    (n, p) = taille(A)
    for k in range(p):
        A[i][k] = a * A[i][k] + b * A[j][k]

def dilate(A, a, i):
    (n, p) = taille(A)
    for k in range(p):
        A[i][k] = a * A[i][k]

#%% exercice 12

# peu de commentaires : c'est une fonction pour chercher un coefficient non-nul
def cherche_pivot(A, i, j):
    (n, p) = taille(A)
    for k in range(i, n):
        if A[k][j] != 0:
            return k
    return None

#%% exercice 13

def echelonne(A):
    (n, p) = taille(A)
    # copie de A
    B = [[A[i][j] for j in range(p)] for i in range(n)]
    # indice du pivot actuel
    r = 0
    for j in range(p):
        k = cherche_pivot(B, r, j)
        # si on a trouvé un pivot : opérations comme d'habitude
        if k != None:
            echange(B, r, k)
            a = B[r][j]
            for t in range(r+1, n):
                b = B[t][j]
                combinaison(B, a, t, -b, r)
            # passer au pivot suivant
            r = r + 1
        # sinon la colonne est nulle en dessous de la ligne r
        # on passe simplement à la colonne suivante (sans incrémenter r)
    print("Rang", r)
    return B

# test
A1 = [[1, 0, -1], [-2, -1, 4], [0, 1, -3]]
matrice_print(A1)
matrice_print(echelonne(A1))
print()
A2 = [[-1, 3, -1], [0, -2, 0], [2, -6, 2], [1, -4, 1]]
matrice_print(A2)
matrice_print(echelonne(A2))
print()
A3 = [[-1, -3, -4, -2, 4], [-1, 2, 4, 5, -1], [-1, -1, -1, 1, 2], [0, 0, 5, -5, 0], [-1, 0, 4, -1, 1], [1, 1, -1, 1, -2]]
matrice_print(A3)
matrice_print(echelonne(A3))

#%% exercice 14

def inverse(A):
    (n, p) = taille(A)
    assert n == p, "pffffffffff"
    # B est une copie de A et I est l'identité
    B = [[A[i][j] for j in range(n)] for i in range(n)]
    I = identité(n)
    r = 0
    # tout pareil au début, mais on opère sur B et sur I en même temps
    for j in range(n):
        k = cherche_pivot(B, r, j)
        # on *doit* trouver un pivot dans chaque colonne, sinon...
        assert k != None, "La matrice n'est pas inversible"
        echange(B, r, k)
        echange(I, r, k)
        a = B[r][j]
        for t in range(r+1, n):
            b = B[t][j]
            combinaison(B, a, t, -b, r)
            combinaison(I, a, t, -b, r)
        r = r + 1
    # ici la matrice est échelonnée et inversible
    # on remonte, en s'intéressant à la ligne n-i-1 pour i de 0 à n-1
    for i in range(n):
        a = B[n-1-i][n-1-i]
        # éliminer, dans la colonne i, les coefficients au dessus de a
        for t in range(0, n-i-1):
            b = B[t][n-1-i]
            combinaison(B, a, t, -b, n-i-1)
            combinaison(I, a, t, -b, n-i-1)
        # puis mettre à 1 les coefficients diagonaux de B
        dilate(B, 1/a, n-1-i)
        dilate(I, 1/a, n-1-i)
    # pour faire des tests : renvoyer B aussi, et vérifier que B est l'identité
    # ceci peut ne pas être exact à cause des problèmes d'arrondi en flottants
    return I

# test avec une matrice inversible (3, 3)
A1 = [[1, 0, -1], [-2, -1, 4], [0, 1, -3]]
matrice_print(A1)
I = inverse(A1)
matrice_print(I)
print("Vérification :")
matrice_print(produit(A1, I))
matrice_print(produit(I, A1))

#%%

# un autre exemple : l'inverse est facile à donner à la main
A5 = diagonale([1, 4, 10])
matrice_print(A3)
matrice_print(inverse(A3))

#%%

# encore un cas : (4, 4), rang 3, non inversible (voir la forme échelonnée)
A6 = [[-3, 1, 1, 1], [1, 2, 0, -4], [-4, 0, 1, 3], [-1, -2, 0, 4]]
matrice_print(A4)
matrice_print(echelonne(A4))
matrice_print(inverse(A4))
